import time, os, sys, shutil
from argparse import ArgumentParser, Namespace, RawDescriptionHelpFormatter

from tipp3 import get_logger, __version__
from tipp3.configs import Configs, _root_dir, main_config_path, _read_config_file
from tipp3.configs import *
from tipp3.refpkg_loader import loadReferencePackage, downloadReferencePackage
from tipp3.query_binning import queryBinning 
from tipp3.query_alignment import queryAlignment
from tipp3.query_placement import queryPlacement
from tipp3.query_abundance import queryAbundance 
from tipp3.helpers.general_tools import *

from multiprocessing import Lock, Manager
from concurrent.futures import ProcessPoolExecutor

_LOG = get_logger(__name__)

global levels, character_map, taxon_map, level_map, key_map
character_map = {'A': 'T', 'a': 't', 'C': 'G', 'c': 'g', 'T': 'A',
                 't': 'a', 'G': 'C', 'g': 'c', '-': '-'}
levels = ["species", "genus", "family", "order",
            "class", "phylum", "superkingdom"]
# available subcommands. If nothing is provided, will default to 'abundance' 
subcommands = ['abundance', 'download_refpkg']

'''
Preset mode to run TIPP3 for abundance profiling
'''
def run_tipp3():
    tipp3_pipeline(mode='tipp3', subcommand='abundance')

'''
Preset mode to run TIPP3-fast for abundance profiling
'''
def run_tipp3_fast():
    tipp3_pipeline(mode='tipp3-fast', subcommand='abundance')

# Main pipeline of TIPP3
def tipp3_pipeline(*args, **kwargs):
    s1 = time.time()
    m = Manager()
    lock = m.Lock()

    # (pre) parse arguments and build configurations
    parser, cmdline_args = parseArguments(
            mode=kwargs.get('mode', None),
            subcommand=kwargs.get('subcommand', None))

    if Configs.command == 'download_refpkg':
        downloadReferencePackage(Configs.outdir, Configs.decompress)
    elif Configs.command == 'abundance':
        # initialize ProcessPoolExecutor
        _LOG.warning('Initializing ProcessorPoolExecutor instance...')
        pool = ProcessPoolExecutor(Configs.num_cpus,
                initializer=initiate_pool, initargs=(parser, cmdline_args,))
        
        # create output directory
        if not os.path.exists(Configs.outdir):
            os.makedirs(Configs.outdir)

        # (0) load refpkg
        refpkg = loadReferencePackage(Configs.refpkg_path, Configs.refpkg_version)

        # (1) read binning against the TIPP3 refpkg using BLAST
        query_paths, query_alignment_paths = queryBinning(refpkg)
        s2 = time.time()
        _LOG.info(f"Runtime for mapping reads to marker genes (seconds): {s2 - s1}") 

        # (2) read alignment to corresponding marker genes
        if Configs.alignment_method != 'blast':
            query_alignment_paths = queryAlignment(refpkg, query_paths)
        s3 = time.time()
        _LOG.info(f"Runtime for aligning reads to marker genes (seconds): {s3 - s2}") 

        # early stop --> alignment-only 
        if Configs.alignment_only:
            _LOG.warning("User specifies to output query alignment to marker "
                    "genes only. Stopping TIPP3 now.")
            _LOG.warning("You can find the alignment files at: "
                    f"{os.path.join(Configs.outdir, 'query_alignments')}")
            tipp3_stop(s1)

        # (3) read placement to corresponding marker gene taxonomic trees
        query_placement_paths = queryPlacement(refpkg, query_alignment_paths)
        s4 = time.time()
        _LOG.info(f"Runtime for placing reads to marker gene taxonomies (seconds): {s4 - s3}") 

        # (4) collect results and abundance profile
        queryAbundance(refpkg, query_placement_paths, pool, lock)
        s5 = time.time()
        _LOG.info(f"Runtime for obtaining abundance profile (seconds): {s5 - s4}") 

        # close ProcessPoolExecutor
        _LOG.warning('Closing ProcessPoolExecutor instance...')
        pool.shutdown()
        _LOG.warning('ProcessPoolExecutor instance closed.')

        # cleaning up temporary files
        if not Configs.keeptemp:
            _LOG.info("Removing intermediate output files...")
            tipp3_clean_temp()
            s6 = time.time()
            _LOG.info(f"Runtime for cleaning temporary files (seconds): {s6 - s5}") 

    # stop TIPP3
    tipp3_stop(s1)

def tipp3_clean_temp():
    temp_dirs = ['blast_output', 'query', 'query_alignments',
            'query_placements']#, 'query_classifications']
    for temp in temp_dirs:
        shutil.rmtree(os.path.join(Configs.outdir, temp))
        _LOG.info(f"Removed {temp}")

def tipp3_stop(start_time):
    send = time.time()
    _LOG.info('TIPP3 completed in {} seconds...'.format(send - start_time))
    #print('TIPP3 completed in {} seconds...'.format(s2 - start_time))
    exit(0)

'''
Init function for a queue and get configurations for each worker
'''
def initiate_pool(parser, cmdline_args):
    buildConfigs(parser, cmdline_args, child_process=True)

'''
parse argument and populate Configs
'''
def parseArguments(mode=None, subcommand=None):
    global _root_dir, main_config_path
    parser = _init_parser(mode)

    cmdline_args = sys.argv[1:]
    # check if subcommand is provided, if not try to append a subcommand
    # if the default is given (e.g., running the binary `tipp3` or
    # `tipp3-accurate`
    if cmdline_args[0] not in subcommands:
        # check if it is an argument or an invalid subcommand
        if cmdline_args[0].startswith('-') and subcommand is not None:
            cmdline_args = [subcommand] + cmdline_args

    buildConfigs(parser, cmdline_args)
    #_LOG = get_logger(__name__, log_path=Configs.log_path)

    getConfigs(arguments=cmdline_args)
    _LOG.info('TIPP3 is running with: {}'.format(' '.join(cmdline_args)))
    #if os.path.exists(main_config_path):
    #    _LOG.info('Main configuration loaded from {}'.format(
    #        main_config_path))
    return parser, cmdline_args

'''
initialize parser to read user inputs
'''
def _init_parser(mode=None):
    # example usages
    example_usages = '''Example usages:
> abundance: run default for profiling (TIPP3-fast)
    %(prog)s abundance -r refpkg_dir/ -i queries.fasta[.gz]
    %(prog)s abundance -r refpkg_dir/ -i queries.fq[.gz]
> abundance: Only output read alignment to marker genes (then exit) 
    %(prog)s abundance -r refpkg_dir/ -i queries.fasta[.gz] --alignment-only
> download_refpkg: Download the latest TIPP3 reference package to current directory and decompress
    %(prog)s download_refpkg -d ./ --decompress
'''

    # determine which mode we have by default (default to tipp3-fast)
    _mode = 'tipp3-fast'
    if mode is not None:
        _mode = mode

    parser = ArgumentParser(
            description=(
                "This program runs TIPP3, a taxonomic identification "
                "and abundance profiling tool for metagenomic reads."),
            conflict_handler='resolve',
            epilog=example_usages,
            #formatter_class=RawDescriptionHelpFormatter)
            formatter_class=SmartHelpFormatter)
    parser.add_argument('-v', '--version', action='version',
        version="%(prog)s " + __version__)

    # add sub-commands
    subparsers = parser.add_subparsers(dest='command', help=None)

    # (1) DEFAULT abundance -- abundance profiling
    subparser_abs = subparsers.add_parser('abundance',
            help="(Default) Abundance profiling on input reads.")

    # (2) download_refpkg -- download reference package to target directory
    subparser_refpkg = subparsers.add_parser('download_refpkg',
            help="Download the latest TIPP3 reference package.")

    # (3) TBD

########################## subcommand: abundance ##############################
    # basic settings
    basic_group = subparser_abs.add_argument_group(
            "Basic parameters".upper(),
            ("These are basic fields for running TIPP3. "
             "Users need to provide the path to a TIPP3-compatible refpkg "
             "and the path to the query reads they wish to profile."))
    subparser_abs.groups = dict()
    subparser_abs.groups['basic_group'] = basic_group
    basic_group.add_argument('-i', '--query-path', type=str,
        help=' '.join(['Path to a set of unaligned query reads',
            'for classification.', 'Accepted format:'
            '.fa/.fasta/.fq/.fastq (can be compressed as a .gz file).']),
        required=True)
    #TODO - not making this required and invoke download for TIPP3 refpkg
    # if no refpkg is supplied
    basic_group.add_argument('-r', '--refpkg-path', '--refpkg',
            '--reference-package',
        type=str, help=' '.join(['Path to a TIPP3-compatible refpkg.',
            'Use subcommand \'download_refpkg\' to download the latest',
            'TIPP3 reference package.']),
        required=False, default=None)
    basic_group.add_argument('--refpkg-version',
        type=str, help='Version of the refpkg. [default: markers-v4]',
        default='markers-v4', required=False)
    basic_group.add_argument('--mode',
        type=str, choices=['tipp3', 'tipp3-fast'], default=mode,
        help=' '.join(['Preset mode for running TIPP3.', f'[default: {_mode}]',
            '\n\"tipp3\": the most accurate setting, with WITCH alignment',
            'and pplacer placement.',
            '\n\"tipp3-fast\": the fastest setting, with BLAST alignment',
            'and Batch-SCAMPP placement.',
            '\nThe mode will be overridden by parameters --alignment-method',
            'and --placement-method.']), required=False)
    basic_group.add_argument('--alignment-method',
        type=str, choices=['witch', 'blast', 'hmm'], default=None,
        help=' '.join(['Alignment method to use for aligning reads',
            'to marker genes. [default: using --mode]']),
        required=False)
    basic_group.add_argument('--placement-method',
        type=str, choices=['pplacer-taxtastic', 'bscampp'], default=None,
        help=' '.join(['Placement method to use for placing aligned reads',
            'to marker gene taxonomic trees. [default: using --mode]']),
        required=False)
    basic_group.add_argument('-c', '--config-file',
        type=str, help=' '.join(['Path to a user-defined config file.',
            'see an example at\n',
            'https://github.com/c5shen/TIPP3/blob/main/custom.config',
            '[default: None]']),
        required=False, default=None)
    basic_group.add_argument('-d', '--outdir',
        type=str, help='Path to the desired output directory [default: ./tipp3_output]',
        required=False, default='./tipp3_output')
    basic_group.add_argument('-t', '--num-cpus',
        type=int, help='Number of CPUs for multi-processing. [default: -1 (all)]',
        required=False, default=-1)

    # miscellaneous group
    misc_group = subparser_abs.add_argument_group(
            "Miscellaneous options".upper(),
            ("Optional parameters for TIPP3 setup/config etc."))
    subparser_abs.groups['misc_group'] = misc_group
    #misc_group.add_argument('--verbose', type=int,
    #        help=' '.join(["Verbose level for logging.",
    #        "0: error, 1: info, >1: debug. [default: 1]"]),
    #        default=1, required=False)
    misc_group.add_argument('--alignment-only', action='store_const',
            const=True, default=False,
            help='Only obtain query alignments to marker genes and stop TIPP3.')
    misc_group.add_argument('--keeptemp', action='store_const', const=True,
            help='Keep temporary files in the running process.',
            default=False)
    misc_group.add_argument('-y', '--bypass-setup', action='store_const',
            const=True, default=True,
            help=' '.join(['(Optional) Include this argument to bypass',
                'the initial step when running TIPP3 to set up the',
                'configuration directory (will use ~/.tipp3).',
                'Note: By default this option is enabled.']))

######################### subcommand: download_refpkg #########################
    refpkg_basic_group = subparser_refpkg.add_argument_group(
            "Basic parameters".upper(),
            ("Use this subcommand to download the latest TIPP3 reference "
             "package, and optionally decompress it for usage."))
    refpkg_basic_group.add_argument('-d', '--outdir',
            type=str, help=' '.join(['Path to put the downloaded refpkg file.',
                '[default: ./]']), default='./')
    refpkg_basic_group.add_argument('--decompress', action='store_const',
            const=True, help='After download, decompress the file for usage.',
            default=False)

    return parser



import os
import zipfile
from tipp3.configs import Configs
from tipp3 import get_logger, log_exception

_LOG = get_logger(__name__)

'''
Load TIPP3 reference package in
'''
def loadReferencePackage(refpkg_path, refpkg_version):
    refpkg = {}

    # sanity check for the existence of refpkg_path
    if not refpkg_path or not os.path.exists(refpkg_path):
        errmsg = ('Refpkg does not exist: {}'.format(refpkg_path) + 
            '\nPlease download reference package using subcommand ' +
            '\"download_refpkg\"')
        _LOG.error(errmsg)
        raise ValueError(errmsg)

    # refpkg dir path from commandline
    path = os.path.join(refpkg_path, refpkg_version)
    input = os.path.join(path, "file-map-for-tipp.txt")
    _LOG.info('Reading refpkg from {}'.format(path))

    # load exclusion list, if any
    exclusion = set() 
    try:
        raw = getattr(Configs, 'refpkg').exclusion
        exclusion = set(raw.strip().split(','))
    except AttributeError:
        pass

    refpkg["genes"] = []
    with open(input) as f:
        for line in f.readlines():
            [key, val] = line.split('=')

            [key1, key2] = key.strip().split(':')

            # hotfix before pushing a new version of TIPP3 refpkg
            # --> change all "taxonomy.table" to "all_taxon.taxonomy"
            if val == 'taxonomy.table':
                val = 'all_taxon.taxonomy'
            val = os.path.join(path, val.strip())

            try:
                refpkg[key1][key2] = val
            except KeyError:
                refpkg[key1] = {}
                refpkg[key1][key2] = val

            if (key1 != "blast") and (key1 != "taxonomy"):
                refpkg["genes"].append(key1)
    
    # add path variable to each marker gene refpkg
    # to use with pplacer-taxtastic
    for marker in refpkg["genes"]:
        marker_refpkg_path = os.path.join(path, f"{marker}.refpkg")
        refpkg[marker]['path'] = marker_refpkg_path

    # excluding marker genes if specified
    _LOG.info('Excluding markers (if exist): {}'.format(exclusion))
    refpkg["genes"] = set(refpkg["genes"]).difference(exclusion)
    refpkg["genes"] = list(refpkg["genes"])
    _LOG.info('Marker genes: {}'.format(refpkg["genes"]))
    _LOG.info('Number of marker genes: {}'.format(len(refpkg["genes"])))

    return refpkg

'''
Download the latest TIPP reference package
'''
def downloadReferencePackage(outdir, decompress=False):
    # hosted on Illinois Databank
    url = 'https://databank.illinois.edu/datafiles/yq7ry/download'
    _LOG.info(f"Downloading the latest TIPP reference package from {url}")

    # specify the name of the downloaded zipfile
    outpath = os.path.join(outdir, 'tipp3-refpkg.zip')
    if os.path.exists(outpath):
        _LOG.warning(
                f"{outpath} exists! Skipped download to avoid overwriting.")
    else:
        cmd = f"wget {url} -O {outpath}"
        os.system(cmd)

    # identify the extract directory name
    try:
        zf = zipfile.ZipFile(outpath)
    except FileNotFoundError as e:
        log_exception(_LOG)
    files = zf.namelist()
    dir = files[0].split('/')[0]

    # decompress if specified
    if decompress:
        _LOG.info(f"Decompressing {os.path.basename(outpath)} to {outdir}")
        if os.path.isdir(dir):
            _LOG.warning(f"\"{dir}/\" from the zipfile exists at {outdir}! "
                    "Skipped decompression to avoid overwriting.")
        else:
            cmd = f"unzip -d {outdir} {outpath}"
            os.system(cmd)
            _LOG.info(f"Decompressed refpkg directory name: {dir}")
    else:
        _LOG.info("Finished downloading. To use with TIPP3, please "
            f"decompress the downloaded file at {outpath}")
    return True



#from dendropy import *
import numpy as np
import heapq
import treeswift
import itertools
from collections import deque
from os.path import expanduser,isfile
import random
import statistics
import copy
import gzip

# store bracket open/close for convenience in label parsing
BRACKET = {
    '[': ']', # square bracket
    '{': '}', # curly bracket
    "'": "'", # single-quote
    '"': '"', # double-quote
}


#separete the query and ref sequence from the alignment file

def read_data(aln):
    """ Load the query and reference sequence from the alignment file
    
    Parameters
    ----------
    aln : multiple sequence alignment containing reference taxa and query sequence
    
    Returns
    -------
    dictionary containing sequences with taxon label keys
    
    """

    f = open(aln)
    result = dict()

    taxa = ""
    seq = ""
    for line in f:
        if line[0] == '>':
            if taxa != "":
                result[taxa] = seq
            taxa = line[1:-1]
            seq = ""

        elif line == "/n":
            continue
        else:
            seq += line[:-1]

    if taxa != "":
        result[taxa] = seq


    return result

def seperate(aln_dict, leaf_dict):
    """ Separate the query sequences from the reference sequences
    
    Parameters
    ----------
    aln_dict : Sequence dictionary with taxon label keys
    leaf_dict : Sequence dictionary with leaf label keys (queries are not in backbone tree)
    
    Returns
    -------
    separate dictionaries containing query sequences and referece sequences with taxon label keys
    
    """
    ref = dict()
    query = dict()

    for key, value in aln_dict.items():
        if key not in leaf_dict:
            query[key] = value
        else:
            ref[key] = value

    return ref, query

def hamming(seq1, seq2):
    """ Returns hamming distance between two sequences
    
    Parameters
    ----------
    seq1 : query sequence
    seq2 : reference sequence
    
    Returns
    -------
    integer hamming distance between query sequence and reference sequence
    
    """
    return sum(1 for ch1, ch2 in zip(seq1, seq2) if ch1 != ch2)


def find_y(x,ref):
    """ Returns leaf label for closest sister taxon l (no longer used)
    
    Parameters
    ----------
    x : aligned query sequence
    ref : reference multiple sequence alignment dictionary 
    
    Returns
    -------
    leaf label for taxon with smallest hamming distacne to query sequence
    
    """
    low = len(x)
    y = ""
    for name, seq in ref.items():
        h_dist = hamming(x, seq)
        if h_dist < low:
            low = h_dist
            y = name
    return y

def set_hamming_dict(args):
    dict_items, other_args = args
    name, seq = dict_items
    ref, n, fragment_flag, y_dict = other_args
    
    y_dict[name] = find_closest_hamming(seq, ref, n, fragment_flag)


def find_closest_hamming(x, ref, n, fragment_flag):
    ''' Returns leaf name for n closest sister taxa to sequence x
    
    Parameters
    ----------
    x : aligned query sequence
    ref : reference multiple sequence alignment dictionary 
    n : number of nodes to return 
    fragment_flag : True if the query is not full length
    
    Returns
    -------
    list of nodes with n smallest hamming distacnes to query sequence

    '''
    queue = []
    closest = []

    counter = 0
    if fragment_flag == 'True':
        [si, ei] = set_fragment_indicies(x)
    else:
        [si, ei] = [0, len(x)]
    
    c = 200 # size of the subtring compared at once
        
    for name, seq in ref.items():
        heapq.heappush(queue,(hamming(seq[si:si+c],x[si:si+c]), ei - si - c, counter, name))
        counter += 1

    while queue:
        (ham_dist, sites_left, cnt, name) = heapq.heappop(queue)
        if sites_left < 0:
            closest.append(name)
            if len(closest) >= n:
                return closest
        else:
            ind = ei - sites_left
            new_ham = hamming(ref[name][ind:ind+c],x[ind:ind+c])
            heapq.heappush(queue,(ham_dist + new_ham, sites_left - c, cnt, name))

def set_fragment_indicies(x):
    """ Returns the indicees without leading and trailing gaps.
    
    Parameters
    ----------
    x = string sequence
    
    Returns
    -------
    list of start index and end index with the first and last non gap character
    
    """
    e = len(x)
    ei = e
    si = 0
    for i in range(ei): 
        if x[i] == '-' and si == i:
            si = i + 1
        if x[e - i - 1] == '-' and ei == e - i:
            ei = e - i - 1
        if ei == si:
            break
    return [si, ei] 

def find_closest(x, visited, y=None):
    """ Returns leaf label for closest leaf to the node x through path not travelling through visited.
    If y is populated returns path from x to y not travelling through nodes in visited.
    
    Parameters
    ----------
    x : dendropy node object
    visited : dictionary containing dendropy node objects as keys 
    y : dendropy node object
    
    Returns
    -------
    If y == None : dendropy node object of closest leaf y to the node x through path not travelling through nodes in visited, 
                   list containing dendropy node objects on path to that leaf y from node x
    If y != None : dendropy node object y, 
                   list containing dendropy node objects on path from node x to leaf y not travelling through nodes in visited
    
    """
    queue = []
    cnt = 1
    visited.add(x)

    if x.get_parent() and x.get_parent() not in visited:
        tmp = []
        tmp.append(x)
        heapq.heappush(queue, [x.get_edge_length(), cnt, tmp, x.get_parent()])
        cnt += 1

    for child in x.child_nodes():
        if child and child not in visited:
            tmp = []
            tmp.append(child)
            heapq.heappush(queue, [child.get_edge_length(), cnt, tmp, child])
            cnt += 1

    while len(queue) > 0:
        try:
            [length, _, path, node] = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf():
            if (not y) or node.get_label()==y.get_label():
                return node, path
            else:
                continue

        if node.get_parent() and node.get_parent() not in visited:
            tmp = path.copy()
            tmp.append(node)
            heapq.heappush(queue, [length+node.get_edge_length(), cnt, tmp, node.get_parent()])
            cnt += 1

        for child in node.child_nodes():
            if child and child not in visited:
                tmp = path.copy()
                tmp.append(child)
                heapq.heappush(queue, [length+child.get_edge_length(), cnt, tmp, child])
                cnt += 1

    return x, [x]


def find_closest_testing(x, visited, y=None, valid_leaves=None):
    """ Returns leaf label for closest leaf to the node x through path not travelling through visited.
    If y is populated returns path from x to y not travelling through nodes in visited.
    
    Parameters
    ----------
    x : dendropy node object
    visited : dictionary containing dendropy node objects as keys
    y : dendropy node object
    
    Returns
    -------
    If y == None : dendropy node object of closest leaf y to the node x through path not travelling through nodes in visited, 
                   list containing dendropy node objects on path to that leaf y from node x
                   distance from node x to leaf y
    If y != None : dendropy node object y, 
                   list containing dendropy node objects on path from node x to leaf y not travelling through nodes in visited, 
                   distance from node x to leaf y
    
    """
    queue = []
    cnt = 1
    visited.add(x)

    if x.get_parent() and x.get_parent() not in visited:
        tmp = []
        tmp.append(x)
        heapq.heappush(queue, [x.get_edge_length(), cnt, tmp, x.get_parent()])
        cnt += 1

    for child in x.child_nodes():
        if child and child not in visited:
            tmp = []
            tmp.append(child)
            heapq.heappush(queue, [child.get_edge_length(), cnt, tmp, child])
            cnt += 1

    while len(queue) > 0:
        try:
            [length, _, path, node] = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf():
            if (not y) or node.get_label() == y.get_label():
                if  (valid_leaves != None and node.get_label() not in valid_leaves):
                    continue
                else:
                    return node, path, length
            else:
                continue

        if node.get_parent() and node.get_parent() not in visited:
            tmp = path.copy()
            tmp.append(node)
            heapq.heappush(queue, [length+node.get_edge_length(), cnt, tmp, node.get_parent()])
            cnt += 1

        for child in node.child_nodes():
            if child and child not in visited:
                tmp = path.copy()
                tmp.append(child)
                heapq.heappush(queue, [length+child.get_edge_length(), cnt, tmp, child])
                cnt += 1

    return x, [x], 0

def build_subtrees(sister_taxon_dict, leaf_dict, tree, nbr_subtrees, subtree_size):
    tree_building_taxa = len(list(sister_taxon_dict))
    cluster_dict = dict()
    cluster_index = 0 
    cluster_list = dict()
    
    

    #tmp
    #draw(tree.extract_tree_with(list(sister_taxon_dict)))
    #tmp

    for taxon, _ in sister_taxon_dict.items():
        cluster_dict[taxon] = cluster_index
        cluster_list[cluster_index] = [taxon]
        cluster_index += 1

    queue = []
    for l1, _ in sister_taxon_dict.items():
        node_l2, _, length = find_closest_testing(leaf_dict[l1],{leaf_dict[l1]},valid_leaves=sister_taxon_dict)
        heapq.heappush(queue, [length, l1, node_l2.get_label()])

    while tree_building_taxa > nbr_subtrees:        
        [length, l1, l2] = heapq.heappop(queue)

        idx = cluster_dict[l1]
        old_idx = cluster_dict[l2]

        if idx == old_idx:
            continue

        for leaf in cluster_list[old_idx]:
            cluster_dict[leaf] = idx
        #[cluster_dict[leaf] = idx for leaf in cluster_list[old_idx]]
        cluster_list[idx].extend(cluster_list[old_idx])
        cluster_list.pop(old_idx)

        tree_building_taxa -= 1

        best_leaf = cluster_list[idx][0]
        for leaf in cluster_list[idx]:
            best_leaf = centered_leaf(tree, cluster_list[idx])
        cluster_node_list = [leaf_dict[leaf] for leaf in cluster_list[idx]]
        cluster_set = {*cluster_node_list}
        node_l3, _, length = find_closest_testing(leaf_dict[best_leaf], cluster_set, valid_leaves=sister_taxon_dict)
        heapq.heappush(queue, [length, best_leaf, node_l3.get_label()])

    trees = []
    query_decomp_dict = []

    for idx, cluster in cluster_list.items():
        best_leaf = centered_leaf(tree, cluster)

        subtree_leaves, tree2 = subtree_with_edge_length(tree, leaf_dict[best_leaf], subtree_size)
        tree2 = tree2.extract_tree_with(subtree_leaves)
        trees.append(tree2)
        query_list = []
        for leaf in cluster:
            query_list.extend(sister_taxon_dict[leaf])
        query_decomp_dict.append(query_list)

    return trees, query_decomp_dict


def subtree_nodes(tree, leaf_y, n):
    """ Returns list of length n of leaves closest to sister taxon
    
    Parameters
    ----------
    tree : treeswift tree object
    leaf_y : treeswift node for closest sister taxon
    n = number of taxa contained in subtree
    
    Returns
    -------
    list of taxon labels corresponding to leaves in the subtree
    
    """
    queue = [(0, 0, leaf_y.get_parent())]

    leaves = [leaf_y]
    visited = {leaf_y}

    counter = 1

    while len(leaves) < n:
        try:
            (length, _, node) = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf():
            leaves.append(node)

        adjacent = node.child_nodes()
        if not node.is_root():
            adjacent.append(node.get_parent())

        for neighbor in adjacent:
            if neighbor not in visited:
                heapq.heappush(queue, (length+1, counter, neighbor))
                counter += 1


    result = []
    for item in leaves:
        result.append(item.get_label())

    return result

def subtree_nodes_with_edge_length(tree, leaf_y, n):
    """ Returns list of length n of leaves closest to sister taxon (minimizing edge weights)
    
    Parameters
    ----------
    tree : treeswift tree object
    leaf_y : treeswift node for closest sister taxon
    n = number of taxa contained in subtree
    
    Returns
    -------
    list of taxon labels corresponding to leaves in the subtree
    """
    queue = [(leaf_y.get_edge_length(), leaf_y.get_parent())]

    leaves = [leaf_y]
    visited = {leaf_y}

    while len(leaves) < n:
        try:
            (length, node) = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        if node.is_leaf() and node.get_label() != '':
            leaves.append(node)

        adjacent_nodes = node.child_nodes()
        if not node.is_root():
            adjacent_nodes.append(node.get_parent())

        for neighbor in adjacent_nodes:
            if neighbor not in visited:
                if neighbor == node.get_parent():
                    heapq.heappush(queue, (length+node.get_edge_length(), neighbor))
                else:
                    heapq.heappush(queue, (length+neighbor.get_edge_length(), neighbor))

    result = []
    for item in leaves:
        result.append(item.get_label())

    return result
    
def subtree_with_edge_length(tree, leaf_y, n):
    """ Returns a subtree and list of length n of leaves closest to sister taxon (minimizing edge weights)
    
    Parameters
    ----------
    tree : treeswift tree object
    leaf_y : treeswift node for closest sister taxon
    n = number of taxa contained in subtree
    
    Returns
    -------
    a subtree and a list of taxon labels corresponding to leaves in the subtree
    """
    if leaf_y.get_parent() == None:
        return None, None
    
    subtree = treeswift.Tree(); subtree.root.label = None; subtree.root.edge_length = 0
    subtreeNode = treeswift.Node(label=leaf_y.get_label(), edge_length=0); subtree.root.add_child(subtreeNode)
        
    queue = [(leaf_y.get_edge_length(), leaf_y.get_parent(), leaf_y.get_edge_length(), subtree.root)]

    leaves = [leaf_y]
    visited = {leaf_y}

    while len(leaves) < n:
        try:
            (length, node, n_edge_length, parent) = heapq.heappop(queue)
        except IndexError:
            break

        visited.add(node)
        subtreeNode = treeswift.Node(label=node.get_label(), edge_length=n_edge_length); parent.add_child(subtreeNode)
        if node.is_leaf():
            leaves.append(node)

        adjacent_nodes = node.child_nodes()
        if not node.is_root():
            adjacent_nodes.append(node.get_parent())

        for neighbor in adjacent_nodes:
            if neighbor not in visited:
                if neighbor == node.get_parent():
                    heapq.heappush(queue, (length+node.get_edge_length(), neighbor, node.get_edge_length(), subtreeNode))
                else:
                    heapq.heappush(queue, (length+neighbor.get_edge_length(), neighbor, neighbor.get_edge_length(), subtreeNode))

    result = []
    for item in leaves:
        result.append(item.get_label())
    
    #subtree.deroot()
    subtree.suppress_unifurcations()
    return result, subtree
    
def extract_taxa_from_tree(a_tree, labels, leaf_dict):
    queue = []
    closest = []

    counter = 0
        
    for label in labels: 
        heapq.heappush(queue,(counter,leaf_dict[label]))
        counter += 1

    while queue:
        (cnt, node) = heapq.heappop(queue)
        if node.num_children() == 0:
            parent = node.get_parent()
            if parent != None:
                parent.remove_child(node)
                heapq.heappush(queue,(counter, parent))
    
    a_tree.resolve_polytomies()            
    a_tree.suppress_unifurcations()

    return a_tree
    
def min_tree_extract_disjoint(a_tree, max_size, tmp_leaves):
    labels, t2 = subtree_with_edge_length(a_tree, tmp_leaves[random.choice(list(tmp_leaves))], max_size)
    
    t2 = t2.extract_tree_with(labels)
    t1 = extract_taxa_from_tree(a_tree, labels, tmp_leaves)
    for label in labels:
       del tmp_leaves[label]
    return t1, t2, tmp_leaves

    
def min_tree_extract_non_disjoint(a_tree, max_size, tmp_leaves):
    labels, t2 = subtree_with_edge_length(a_tree, tmp_leaves[random.choice(list(tmp_leaves))], max_size)
    
    t2 = a_tree.extract_tree_with(labels)
    for label in labels:
       if label in tmp_leaves:
           del tmp_leaves[label]
    return a_tree, t2, tmp_leaves
    
def decompose_tree(a_tree,max_size):
    tree_list = []
    tmp_leaves = a_tree.label_to_node(selection='leaves')
    if '' in tmp_leaves:
        del tmp_leaves['']
    #t1, t2, tmp_leaves = min_tree_extract_disjoint(a_tree, max_size, tmp_leaves)
    t1, t2, tmp_leaves = min_tree_extract_non_disjoint(a_tree, max_size, tmp_leaves)
    while len(tmp_leaves) > 0:
        tree_list.append(t2)

        #t1, t2, tmp_leaves = min_tree_extract_disjoint(t1, max_size, tmp_leaves)
        t1, t2, tmp_leaves = min_tree_extract_non_disjoint(t1, max_size, tmp_leaves)
    tree_list.append(t2)
    print ("nbr subtrees: ", len(tree_list))
    return tree_list
    
def decompose_tree_min_clust(a_tree,max_size):
    tree_list = []
    t1, t2 = min_cluster_size_bisect(a_tree,max_size)
    while t2 != None:
        tree_list.append(t2)
        t1, t2 = min_cluster_size_bisect(t1, max_size)
    tree_list.append(t1) #this might be a bug
    return tree_list    
    
def min_cluster_size_bisect(a_tree,max_size):
    '''
    modified from PASTA to use treeswift
    '''
    nleaf = dict()

    print("before extracting subtree: " + str(len(a_tree.label_to_node(selection='leaves'))))
    #a_tree.draw()
    
    for node in a_tree.traverse_postorder():
        if node.is_leaf():
            nleaf[node] = 1
        else:
            nleaf[node] = 0
            max_child = None
            max_nleaf = 0
            for ch in node.child_nodes():
               nleaf[node] += nleaf[ch]
               if nleaf[ch] > max_nleaf:
                   max_nleaf = nleaf[ch]
                   max_child = ch
        if nleaf[node] >= max_size:
            node.remove_child(max_child)
            t1 = a_tree.extract_subtree(max_child)
            print("subtree size:              " + str(len(t1.label_to_node(selection='leaves'))))
            print("after extracting subtree:  " + str(len(a_tree.label_to_node(selection='leaves'))))
            #t1.deroot()
            #t1.draw()
            t1.resolve_polytomies()
            return a_tree,t1
    
    print("after extracting subtree:  " + str(len(a_tree.label_to_node(selection='leaves'))))
    return a_tree,None      

def centered_leaf(tree, cluster):
    #pick best sister leaf for cluster center
    best_leaf = cluster[0]
    if len(cluster) > 1:
        tmp_tree = tree.extract_tree_with(cluster)
        min_distance = 99999999999
        tmp_label_dict = tmp_tree.label_to_node(selection='leaves')
        for leaf in cluster: 
            total = total_distance(tmp_tree, tmp_label_dict[leaf])

            if total < min_distance:
                min_distance = total
                best_leaf = leaf

    return best_leaf

def max_distance(tree, node):
    maximum = 0 
    for leaf in tree.traverse_leaves():
        distance = tree.distance_between(node, leaf)
        if distance >= maximum:
            maximum = distance
    return maximum

def total_distance(tree, node):
    total = 0 
    for leaf in tree.traverse_leaves():
        total += tree.distance_between(node, leaf)
    return total


def avg_distance(tree, node):
    total = 0 
    leaf_cnt = 0
    for leaf in tree.traverse_leaves():
        total += tree.distance_between(node, leaf)
        leaf_cnt += 1
    return total/leaf_cnt

def median_distance(tree, node):
    total = 0
    leaf_cnt = 0
    distances = []
    for leaf in tree.traverse_leaves():
        distance = tree.distance_between(node, leaf)
        total += distance
        distances.append(distance)
        leaf_cnt += 1
    return statistics.median(distances)

def add_edge_nbrs(tree):
    counter = 0
    for node in tree.traverse_postorder():
        #if not node.is_root():
            counter += 1
            label = node.get_label()
            if label == None:
                node.set_label('%%{}'.format(counter))
            else:
                node.set_label('{}%%{}'.format(label, counter))

def remove_edge_nbrs(tree):
    for node in tree.traverse_postorder():
        #if not node.is_root():
            label_list = node.get_label().split('%%',1)
            if label_list[0] == '':
                node.set_label(None)
            else:
                node.set_label(label_list[0])

'''
   The following three functions are modified from treeswift to
   read and write newick files with jplace tokens
'''
def newick_edge_tokens(tree):
    '''
    Modified from treeswift tree.newick()
    Output this ``Tree`` as a Newick string with lables
    Returns:
        ``str``: Newick string of this ``Tree``
    '''
    label_list = tree.root.get_label().split('%%',1)

    if tree.root.edge_length is None:
        suffix = ';'
    elif isinstance(tree.root.edge_length,int):
        suffix = '%s:%d{%d};' % (str(label_list[0]), int(tree.root.edge_length), int(label_list[1]))
    elif isinstance(tree.root.edge_length,float) and tree.root.edge_length.is_integer():
        suffix = '%s:%d{%d};' % (str(label_list[0]), float(tree.root.edge_length), int(label_list[1]))
    else:
        suffix = '%s:%s{%d};' % (str(label_list[0]), str(tree.root.edge_length), int(label_list[1]))

    if tree.is_rooted:
        return '[&R] %s%s' % (newick_edge_tokens_node(tree.root),suffix)
    else:
        return '%s%s' % (newick_edge_tokens_node(tree.root),suffix)

def newick_edge_tokens_node(node):
    '''
    Modified from treeswift node.newick()
    Newick string conversion starting at this ``Node`` object
    Returns:
        ``str``: Newick string conversion starting at this ``Node`` object
    '''
    node_to_str = dict()
    for node in node.traverse_postorder():
        node_label = node.get_label()
        [label, edge_nbr] = node_label.split('%%',1)
        #node.set_label(label_list[0])
        if node.is_leaf():
            if label is None:
                node_to_str[node] = ''
            else:
                node_to_str[node] = str(label)
        else:
            out = ['(']
            for c in node.children:
                c_label = c.get_label()
                [label_c, edge_nbr_c] = c_label.split('%%',1)
                out.append(node_to_str[c])
                if c.edge_length is not None:
                    if isinstance(c.edge_length,int):
                        l_str = str(c.edge_length)
                    elif isinstance(c.edge_length,float) and c.edge_length.is_integer():
                        l_str = str(int(c.edge_length))
                    else:
                        l_str = str(c.edge_length)
                    out.append(':%s{%d}' % (l_str, int(edge_nbr_c)))
                out.append(',')
                del node_to_str[c]
            out.pop() # trailing comma
            out.append(')')
            if label is not None:
                out.append(str(label))
            node_to_str[node] = ''.join(out)
    return node_to_str[node]

#def write_tree_newick_edge_tokens(tree, filename, hide_rooted_prefix=False):
#    '''
#    Modified from treeswift tree.write_tree_newick()
#    Write this ``Tree`` to a Newick file
#       Args:
#        ``filename`` (``str``): Path to desired output file (plain-text or gzipped)
#    '''
#    if not isinstance(filename, str):
#        raise TypeError("filename must be a str")
#    treestr = newick_edge_nbr_string(tree)
#    if hide_rooted_prefix:
#        if treestr.startswith('[&R]'):
#            treestr = treestr[4:].strip()
#    else:
#        raise Warning("Specified hide_rooted_prefix, but tree was not rooted")
#    if filename.lower().endswith('.gz'): # gzipped file
#        f = gzip.open(expanduser(filename),'wb',9); f.write(treestr.encode()); f.close()
#    else: # plain-text file
#        f = open(expanduser(filename),'w'); f.write(treestr); f.close()

def read_tree_newick_edge_tokens(newick):
    '''
    Modified from treeswift.read_tree_newick(newick)
    Read a tree from a Newick string or file
    Args:
        ``newick`` (``str``): Either a Newick string or the path to a Newick file (plain-text or gzipped)

    Returns:
        ``Tree``: The tree represented by ``newick``. If the Newick file has multiple trees (one per line), a ``list`` of ``Tree`` objects will be returned
    '''
    place_edge_dict = dict()
    if not isinstance(newick, str):
        try:
            newick = str(newick)
        except:
            raise TypeError("newick must be a str")
    if newick.lower().endswith('.gz'): # gzipped file
        f = gzip.open(expanduser(newick)); ts = f.read().decode().strip(); f.close()
    elif isfile(expanduser(newick)): # plain-text file
        f = open(expanduser(newick)); ts = f.read().strip(); f.close()
    else:
        ts = newick.strip()
    lines = ts.splitlines()
    if len(lines) != 1:
        return [read_tree_newick_edge_tokens(l) for l in lines]
    try:
        t = treeswift.Tree(); t.is_rooted = ts.startswith('[&R]')
        if ts[0] == '[':
            ts = ']'.join(ts.split(']')[1:]).strip(); ts = ts.replace(', ',',')
        n = t.root; i = 0
        while i < len(ts):
            # end of Newick string
            if ts[i] == ';':
                if i != len(ts)-1 or n != t.root:
                    raise RuntimeError("Invalid newick file!")

            # go to new child
            elif ts[i] == '(':
                c = treeswift.Node(); n.add_child(c); n = c

            # go to parent
            elif ts[i] == ')':
                n = n.parent

            # go to new sibling
            elif ts[i] == ',':
                n = n.parent; c = treeswift.Node(); n.add_child(c); n = c

            # edge length
            elif ts[i] == ':':
                i += 1; ls = ''
                while ts[i] != ',' and ts[i] != ')' and ts[i] != ';' and ts[i] != '{':
                    ls += ts[i]; i += 1
                if ls[0] == '[':
                    n.edge_params = ']'.join(ls.split(']')[:-1]); ls = ls.split(']')[-1]
                n.edge_length = float(ls); i -= 1

            # edge token
            elif ts[i] == '{':
                i += 1; ls = ''
                while ts[i] != '}':
                    ls += ts[i]; i += 1
                place_edge_dict[ls] = n

            # node label
            else:
                label = ''; bracket = None
                while bracket is not None or ts[i] in BRACKET or (ts[i] != ':' and ts[i] != ',' and ts[i] != ';' and ts[i] != ')'):
                    if ts[i] in BRACKET and bracket is None:
                        bracket = ts[i]
                    elif bracket is not None and ts[i] == BRACKET[bracket]:
                        bracket = None
                    label += ts[i]; i += 1
                i -= 1; n.label = label
            i += 1
    except Exception as e:
        raise RuntimeError("Failed to parse string as Newick")
    return t, place_edge_dict



import os, time
try:
    import configparser
except ImportError:
    from ConfigParser import configparser
from argparse import ArgumentParser, Namespace
from platform import platform
from tipp3.init_configs import init_config_file
from tipp3 import get_logger

# detect home.path location or creating one if it is missing
homepath = os.path.dirname(__file__) + '/home.path'
_root_dir, main_config_path = init_config_file(homepath)

# set of valid configparse section names
valid_config_sections = ['witch', 'bscampp', 'pplacer-taxtastic', 
        'blast', 'refpkg'] 
logging_levels = set(['DEBUG', 'INFO', 'WARNING', 'ERROR', 'CRITICAL'])

_LOG = get_logger(__name__)

# default tqdm style for progress bar
tqdm_styles = {
        'desc': '\tRunning...', 'ascii': False,
        'ncols': 80, 
        #'disable': True,
        #'colour': 'green',
        'mininterval': 0.5
        }

'''
Configurations defined by users
'''
class Configs:
    global _root_dir

    # subcommand and verbose
    command = None
    verbose = 'INFO'

    # Multiprocessing settings
    num_cpus = -1
    max_concurrent_jobs = None

    # basic input items
    query_path = None
    refpkg_path = None     # e.g., xxx/yyy/tipp3-refpkg
    outdir = None
    config_file = None     # Added @ 7.25.2024 

    # choices of parameters
    # default to TIPP3-fast
    mode = 'tipp3-fast'
    alignment_method = 'blast'  # or blast
    placement_method = 'bscampp'  # or other method

    # binary paths (from config file)
    # these are just the default ones. User can modularize the method
    # for each step and add the corresponding executable paths
    # in main.config (by default at ~/.tipp3/main.config)
    pplacer_path = None
    bscampp_path = None
    blastn_path = None
    witch_path = None
    tippjsonmerger_path = None

    # reference package dir path
    refpkg_version = 'markers-v4'


    # miscellaneous
    alignment_only = False
    keeptemp = False
    bypass_setup = True
    
    ########## configs specific for download_refpkg ###########
    decompress = False

# check for valid configurations and set them
def set_valid_configuration(name, conf):
    if not isinstance(conf, Namespace):
        _LOG.warning('Looking for Namespace object from \'{}\' but find {}'.format(
            name, type(conf)))
        return

    # backbone alignment settings
    if name == 'basic':
        for k in conf.__dict__.keys():
            attr = getattr(conf, k)
            if not attr:
                continue

            if k == 'alignment_method':
                assert str(attr).lower() in ['witch', 'blast', 'hmm'], \
                    'Alignment method {} not implemented'.format(attr)
            elif k == 'placement_method':
                pass
                #assert int(attr).lower() in ['pplacer', 'bscampp'], \
                #    'Placement method {} not implemented'.format(attr)
            #elif k == 'path':
            #    assert os.path.exists(os.path.realpath(str(attr))), \
            #        '{} does not exist'.format(os.path.realpath(str(attr)))
            setattr(Configs, k, attr)
    elif name in valid_config_sections:
        setattr(Configs, name, conf)
    else:
        # not reading any invalid sections
        pass

# valid attribute check
def valid_attribute(k, v):
    assert isinstance(k, str)
    if isinstance(v, staticmethod):
        return False
    if not k.startswith('_'):
        return True
    return False

# print a list of all configurations
def getConfigs(arguments=None):
    msg = '\n********** Configurations **********\n' + \
            '\thome.path: {}\n'.format(homepath) + \
            '\tmain.config: {}\n'.format(main_config_path) + \
            '\targuments: {}\n\n'.format(arguments)
    for k, v in Configs.__dict__.items():
        if valid_attribute(k, v):
            msg += '\tConfigs.{}: {}\n'.format(k, v)
    print(msg, flush=True)

'''
Read in from config file if it exists. Any cmd-line provided configs will
override the config file.

Original functionality comes from SEPP -> sepp/config.py
'''
def _read_config_file(filename, cparser, opts,
        child_process=False, expand=None):
    #if not child_process:
    #    _LOG.info('Reading config from {}'.format(filename))
    config_defaults = []

    with open(filename, 'r') as cfile:
        cparser.read_file(cfile)
        if cparser.has_section('commandline'):
            for k, v in cparser.items('commandline'):
                config_defaults.append('--{}'.format(k))
                config_defaults.append(v)

        for section in cparser.sections():
            if section == 'commandline':
                continue
            if getattr(opts, section, None):
                section_name_space = getattr(opts, section)
            else:
                section_name_space = Namespace()
            for k, v in cparser.items(section):
                if expand and k == 'path':
                    v = os.path.join(expand, v)
                section_name_space.__setattr__(k, v)
            opts.__setattr__(section, section_name_space)
    return config_defaults

'''
Validate that Configs is set up correctly, mainly checking if the binary paths
exist
'''
def validateConfigs():
    b_valid = True
    ret = []
    files_to_check = {'PPLACER': Configs.pplacer_path,
            'BSCAMPP': Configs.bscampp_path,
            'BLASTN': Configs.blastn_path,
            'WITCH': Configs.witch_path}
    for method, path in files_to_check.items():
        if (not path) or (path == '') or (not os.path.exists(path)):
            b_valid = False
            ret.append((method, path))
    return b_valid, ret

'''
Build configurations
'''
def buildConfigs(parser, cmdline_args, child_process=False, rerun=False):
    # config parser, which first reads in main.config and later overrides
    # with user.config (if specified)
    cparser = configparser.ConfigParser()
    cparser.optionxform = str

    # load cmdline args first and identify the output directory
    # (to quickly create the outdir and log file
    args = parser.parse_args(cmdline_args)
    Configs.outdir = os.path.realpath(args.outdir)
    #if not os.path.exists(Configs.outdir):
    #    os.makedirs(Configs.outdir)

    # load default_args from main.config
    default_args = Namespace()
    cmdline_default = _read_config_file(main_config_path,
            cparser, default_args, child_process=child_process)
    
    # load cmdline args first, then search for user.config if specified
    args = parser.parse_args(cmdline_args)
    cmdline_user = []
    if 'config_file' in args.__dict__ and args.config_file != None:
        # override default_args
        Configs.config_file = args.config_file
        cmdline_user = _read_config_file(Configs.config_file,
                cparser, default_args, child_process=child_process)

    # finally, re-parse cmdline args in the order:
    #   [cmdline_default, cmd_user, cmdline_args] 
    args = parser.parse_args(cmdline_default + cmdline_user + cmdline_args,
            namespace=default_args)

    # store sub-command given
    Configs.command = args.command

    ######### subcommand: abundance ##########
    if Configs.command == 'abundance':
        # Must have
        Configs.query_path = os.path.realpath(args.query_path)
        if args.refpkg_path:
            Configs.refpkg_path = os.path.realpath(args.refpkg_path)

        Configs.alignment_only = args.alignment_only
        Configs.keeptemp = args.keeptemp

        # set up preset mode, TIPP3 or TIPP3-fast
        Configs.mode = args.mode
        if args.mode == 'tipp3-fast':
            Configs.alignment_method = 'blast'
            Configs.placement_method = 'bscampp'
        elif args.mode == 'tipp3':
            Configs.alignment_method = 'witch'
            Configs.placement_method = 'pplacer-taxtastic'

        # alignment_method and placement_method, and refpkg version
        if args.alignment_method:
            Configs.alignment_method = args.alignment_method
        if args.placement_method:
            Configs.placement_method = args.placement_method
        Configs.refpkg_version = args.refpkg_version

        if args.num_cpus > 0:
            Configs.num_cpus = min(os.cpu_count(), args.num_cpus)
        else:
            Configs.num_cpus = os.cpu_count()

        # verbose level
        verbose = os.getenv('TIPP_LOGGING_LEVEL', 'info').upper()
        if verbose in logging_levels:
            Configs.verbose = verbose
    ############ subcommand: download_refpkg #############
    elif Configs.command == 'download_refpkg':
        Configs.outdir = args.outdir
        Configs.decompress = args.decompress
    else:
        raise NotImplementedError
    
    #if args.max_concurrent_jobs:
    #    Configs.max_concurrent_jobs = args.max_concurrent_jobs
    #else:
    #    Configs.max_concurrent_jobs = min(50, 10 * Configs.num_cpus)

    # add any additional arguments to Configs
    for k in args.__dict__.keys():
        if k not in Configs.__dict__:
            k_attr = getattr(args, k)

            # check whether the configuration is valid
            set_valid_configuration(k, k_attr)

    # try once for validating Configs being set up correctly
    # if not, try once for re-initializing the main config file
    b_valid, invalid_paths = validateConfigs()
    if not b_valid and not rerun:
        _LOG.warning('Some software required by TIPP3 do not have valid binaries!')

        # trying once for regenerating
        _LOG.warning('Re-initializing the config file for once...')
        init_config_file(homepath, rerun=True)
        buildConfigs(parser, cmdline_args, rerun=True)
    elif not b_valid and rerun:
        # if re-initializing does not fix the problem
        if not b_valid:
            errmsg = 'Failed to find valid binaries for some software:\n'
            for item in invalid_paths:
                errmsg += f'\t{item[0]}: {item[1]}\n'
                _LOG.error(errmsg)
                raise ValueError(errmsg)



#!/usr/bin/env python3

'''
This version uses a new technique for voting and ensuring trees contain all of the query sequences votes!
'''

import sys
import os
import utils
import shutil
import json
import time
import argparse
import treeswift
import copy
import multiprocessing as mp
import threading
import itertools
from collections import Counter

__root = os.path.dirname(os.path.realpath(__file__))

def main(args):
    tree_path = args.tree
    output = args.outdir
    outFile = args.output
    aln = args.alignment
    n = args.subtreesize
    run = args.tmpfilenbr
    similarity_flag = args.similarityflag
    fragment_flag = args.fragmentflag
    q_aln = args.qalignment
    model = args.model
    info = args.info
    nbr_closest = args.votes
    num_threads = args.threads

    # filter-min and filter-max
    filter_min, filter_max = args.filter_min, args.filter_max
    filter_acc_lwr = args.filter_acc_lwr

    # output path, ref, query, backbone tree, info
    t0 = time.perf_counter()
    tree = treeswift.read_tree_newick(tree_path)
    tree.resolve_polytomies()

    leaf_dict = tree.label_to_node(selection='leaves')
    # clean the leaf keys so that ' or " are not present
    ori_keys = list(leaf_dict.keys())
    for key in ori_keys:
        _node = leaf_dict[key]
        new_key = key.replace('\'', '')
        new_key = new_key.replace('\"', '')
        leaf_dict.pop(key)
        leaf_dict[new_key] = _node
    #print(len(leaf_dict), ori_keys, leaf_dict)
    
    print ('{} seconds loading tree'.format(time.perf_counter() - t0))

    try:
        os.mkdir("{}/tmp{}".format(output, run))
    except OSError as error:
    	pass
    try:
        os.mkdir(output)
    except OSError as error:
        pass

    if q_aln != "":
        ref_dict = utils.read_data(aln)
        q_dict = utils.read_data(q_aln)
    else:
        aln_dict = utils.read_data(aln)
        ref_dict, q_dict = utils.seperate(aln_dict, leaf_dict)
        #print(len(ref_dict), len(q_dict))
        #exit()

        q_aln = "{}/tmp{}/".format(output, run) + "qaln.fa"
        write_fasta(q_aln, q_dict)

        aln = "{}/tmp{}/".format(output, run) + "aln.fa"
        write_fasta(aln, ref_dict)

    print ('{} seconds loading alignment'.format(time.perf_counter() - t0))

    query_votes_dict = dict()
    query_top_vote_dict = dict()

    tmp_output = "{}/tmp{}/".format(output, run) + "/closest.txt"

    if similarity_flag == True:
        os.system("{}/homology {} {} {} {} {} {}".format(__root, aln, len(ref_dict), q_aln, len(q_dict), tmp_output, nbr_closest))
    else:
        if fragment_flag == False:
            os.system("{}/hamming {} {} {} {} {} {}".format(__root, aln, len(ref_dict), q_aln, len(q_dict), tmp_output, nbr_closest))
        else: 
            os.system("{}/fragment_hamming {} {} {} {} {} {}".format(__root, aln, len(ref_dict), q_aln, len(q_dict), tmp_output, nbr_closest))

    print ('{} seconds finding closest leaves'.format(time.perf_counter() - t0))
    
    unusable_queries = set()
    f = open(tmp_output)
    for line in f:
        line = line.strip()
        y = line.split(',')
        name = y.pop(0)
        #print(y)
        for idx, taxon in enumerate(y):

            leaf, hamming = taxon.split(':')
            y[idx] = (leaf, int(hamming))

        y = sorted(y, key=lambda x: x[1])
        for idx, taxon in enumerate(y):
            y[idx] = taxon[0]

        if name.find(':') >= 0:
            name_list = name.split(":")
            name = name_list[0]
            ungapped_length = name_list[1]
            if y[0] == ungapped_length:
                print ('Error with sequence', name, ': No homologous sites found in the reference sequences. Query removed before placement.')
                unusable_queries.add(name)
        if name not in unusable_queries:
            query_votes_dict[name] = y
            query_top_vote_dict[name] = y[0]
    f.close()

    print ('{} seconds processing closest leaves'.format(time.perf_counter() - t0))

    lf_votes = Counter()
    leaf_queries = dict()
    for name, y in query_votes_dict.items():

        lf_votes.update(y)
        for ind, leaf in enumerate(y):
            top_vote = False
            if ind == 0:
                top_vote = True
            if leaf not in leaf_queries:           
                leaf_queries[leaf] = {(name,top_vote)}
            else:
                leaf_queries[leaf].add((name,top_vote))

    subtree_dict = dict()
    subtree_leaf_label_dict = dict()
    most_common_index = 0
    
    while len(query_votes_dict) > 0:
        #print(lf_votes.most_common(most_common_index+1), most_common_index)
        (seed_label, node_votes) = lf_votes.most_common(most_common_index+1)[most_common_index]
        
        node_y = leaf_dict[seed_label]
        labels = utils.subtree_nodes_with_edge_length(tree, node_y, n)
        subtree = tree.extract_tree_with(labels)
        label_set = set(labels)

        queries_by_subtree = set()
        subtree_query_set = set()

        #gather any other queries that can be used with this subtree
        for label in labels:
            leaf_queries_remove_set = set()
            if label in leaf_queries:
                    
                for leaf_query, top_vote in leaf_queries[label]:
                
                    if leaf_query not in query_votes_dict:
                        leaf_queries_remove_set.add((leaf_query, top_vote))
                        continue
                        
                    if top_vote:
                        subtree_query_set.add(leaf_query)
                        leaf_queries_remove_set.add((leaf_query, top_vote))
                    
                leaf_queries[label].difference_update(leaf_queries_remove_set)
        queries_by_subtree.update(subtree_query_set)

        if len(queries_by_subtree) > 0:
            subtree_dict[subtree] = (seed_label, queries_by_subtree)
            subtree_leaf_label_dict[subtree] = subtree.label_to_node(selection='leaves')

        votes_b4 = len(list(lf_votes.elements()))
        
        for query in queries_by_subtree:
            if query in query_votes_dict:
                lf_votes.subtract(query_votes_dict[query])
                query_votes_dict.pop(query)
        if len(queries_by_subtree)> 0:        
            print ("queries left: {}".format(len(query_votes_dict)))
        if len(queries_by_subtree) == 0:
            most_common_index += 1
            # 10.27.2023 - Chengze Shen
            # >>> prevent going over the the total number of votes
        else:
            most_common_index = 0
            
    jplace = dict()
    placements = []

    utils.add_edge_nbrs(tree)
    jplace["tree"] = utils.newick_edge_tokens(tree)
    print ('{} seconds adding tokens'.format(time.perf_counter() - t0))

    placed_query_list = []
    
    # reassign queries to the subtree minimizing total edge length 
    # from the query's top vote to the subtree's seedleaf
    new_subtree_dict = dict()
    for query, closest_label in query_top_vote_dict.items():
   
        best_subtree = None
        best_distance = 99999999999999999
        for subtree, value in subtree_dict.items():
            leaf_label_dict = subtree_leaf_label_dict[subtree]
            seed_label, _ = value
            if closest_label in leaf_label_dict:
                distance = subtree.distance_between(leaf_label_dict[closest_label], leaf_label_dict[seed_label])
                if distance < best_distance:
                   best_subtree = subtree
                   best_distance = distance
        if best_subtree in new_subtree_dict:
            new_subtree_dict[best_subtree].append(query)
        else:
            new_subtree_dict[best_subtree] = [query]
        
            

    print ('{} seconds assigning subtrees'.format(time.perf_counter() - t0))
    final_subtree_count = 0
    
    tmp_output = "{}/tmp{}/query_subtree_dict".format(output, run)
    
    for subtree, query_list in new_subtree_dict.items():

        if len(query_list) == 0:
            continue

        final_subtree_count += 1
        
        tmp_tree = "{}/tmp{}/tree".format(output, run)
        tmp_aln = "{}/tmp{}/aln".format(output, run) + ".fa"
        tmp_qaln = "{}/tmp{}/qaln".format(output, run) + "q.fa"
        tmp_output = "{}/tmp{}/".format(output, run) + "/epa_result.jplace"
        tmp_dir = "{}/tmp{}/".format(output, run)
        try:
            os.mkdir(tmp_dir)
        except OSError as error:
            pass

        tmp_leaf_dict = subtree.label_to_node(selection='leaves')
        if '' in tmp_leaf_dict:
            del tmp_leaf_dict['']
        tmp_ref_dict = {label : ref_dict[label] for label in tmp_leaf_dict.keys()}
        write_fasta(tmp_aln, tmp_ref_dict)
        
        tmp_q_dict = {name : q_dict[name] for name in query_list}
        write_fasta(tmp_qaln, tmp_q_dict)

        #print ('{} seconds building alignment'.format(time.perf_counter() - t0))

        subtree.resolve_polytomies()
        subtree.suppress_unifurcations()
        subtree.write_tree_newick(tmp_tree, hide_rooted_prefix=True)

        #print ('{} seconds writing subtree'.format(time.perf_counter() - t0))

        cmd = "{}/epa-ng -m {} -t {} -w {} -s {} -q {} --redo -T {} --filter-min {} --filter-max {}".format(
            __root, info, tmp_tree, tmp_dir, tmp_aln, tmp_qaln, num_threads,
            filter_min, filter_max)
        if filter_acc_lwr != None:
            cmd += ' --filter-acc-lwr {}'.format(filter_acc_lwr)
        os.system(cmd)

        #print ('{} seconds running epa-ng'.format(time.perf_counter() - t0))

        place_file = open(tmp_output, 'r')
        place_json = json.load(place_file)
        
        
        if len(place_json["placements"]) > 0:

            added_tree, edge_dict = utils.read_tree_newick_edge_tokens(place_json["tree"])

            for tmp_place in place_json["placements"]:
                placed_query_list.append(tmp_place["n"][0])
                for i in range(len(tmp_place["p"])):

                    edge_num = tmp_place["p"][i][0]
                    edge_distal = tmp_place["p"][i][3]

                    right_n = edge_dict[str(edge_num)]
                    left_n = right_n.get_parent()

                    #left and right path_l and path_r are in added_tree
                    left, path_l = utils.find_closest(left_n, {left_n, right_n})
                    right, path_r = utils.find_closest(right_n, {left_n, right_n})

                    left = leaf_dict[left.get_label()]
                    right = leaf_dict[right.get_label()]
                    _, path = utils.find_closest(left, {left}, y=right)
                    # now left right and path are in tree

                    length = sum([x.get_edge_length() for x in path_l])+edge_distal
                    # left path length through subtree before placement node

                    target_edge = path[-1]

                    for j in range(len(path)):
                        length -= path[j].get_edge_length()
                        if length < 0:
                            target_edge = path[j]
                            break

                    tmp_place["p"][i][0] = 0

                    label = target_edge.get_label()

                    [taxon, target_edge_nbr] = label.split('%%',1)
                    tmp_place["p"][i][0] = target_edge.get_edge_length()+length
                    tmp_place["p"][i][1] = int(target_edge_nbr)

                placements.append(tmp_place.copy())

        place_file.close()
    
    jplace["placements"] = placements
    jplace["metadata"] = {"invocation": " ".join(sys.argv)}
    jplace["version"] = 3
    jplace["fields"] = ["distal_length", "edge_num", "like_weight_ratio", \
            "likelihood", "pendant_length"]


    output = open('{}/{}.jplace'.format(output, outFile), 'w')
    json.dump(jplace, output, sort_keys=True , indent=4)
    output.close()
    print ('{} seconds building jplace'.format(time.perf_counter() - t0))
    print ('Final number of subtrees used:', final_subtree_count)
#    shutil.rmtree("tmp{}".format(run))
    

def write_fasta(aln, aln_dict, aligned=True):
        
        f = open(aln, "w")
        for label, seq in aln_dict.items():
            if label != '':
                f.write(">"+label+"\n")
                if aligned:
                    f.write(seq+"\n")
                else:
                    f.write(seq.replace('-','')+"\n")
        f.close()

def parseArgs():
    parser = argparse.ArgumentParser()

    parser.add_argument("-i", "--info", type=str,
                        help="Path to model parameters", required=True, default=None)
    
    parser.add_argument("-t", "--tree", type=str,
                        help="Path to reference tree with estimated branch lengths", required=True, default=None)
    
    parser.add_argument("-d", "--outdir", type=str,
                        help="Directory path for output", required=True, default=None)
    
    parser.add_argument("-a", "--alignment", type=str,
                        help="Path for query and reference sequence alignment in fasta format", required=True, default=None)

    parser.add_argument("-o", "--output", type=str,
                        help="Output file name", required=False, default="EPA-ng-BSCAMPP")

    parser.add_argument("-m", "--model", type=str,
                        help="Model used for edge distances",
                        required=False, default="GTR")

    parser.add_argument("-b", "--subtreesize", type=int,
                        help="Integer size of the subtree",
                        required=False, default=2000)
    
    parser.add_argument("-V", "--votes", type=int,
                        help="Integer number of votes per query sequence",
                        required=False, default=5)
    
    parser.add_argument("-s", "--similarityflag", type=str2bool,
                        help="boolean, False if maximizing sequence similarity instead of simple Hamming distance (ignoring gap sites in the query)",
                        required=False, default=True)
    
    parser.add_argument("-n","--tmpfilenbr", type=int,
                        help="tmp file number",
                        required=False, default=0)
    
    parser.add_argument("-q", "--qalignment", type=str,
                        help="Path to query sequence alignment in fasta format (ref alignment separate)",
                        required=False, default="")
    
    parser.add_argument("-f", "--fragmentflag", type=str2bool,
                        help="boolean, True if queries contain fragments",
                        required=False, default=True)
    parser.add_argument('--threads', type=int,
            help='number of threads for EPA-ng, default: all',
            required=False, default=-1)
    
    # 11.4.2023 - Chengze Shen
    # added option to set minimum and maximum of reported placements by EPA-ng
    parser.add_argument('--filter-min', type=int,
            help='Minimum number of placements to report, default: 1',
            required=False, default=1)
    parser.add_argument('--filter-max', type=int,
            help='Maximum number of placements to report, default: 7',
            required=False, default=7)
    parser.add_argument('--filter-acc-lwr', type=float,
            help='Accumulated threshold to stop reporting placements, default: None',
            required=False, default=None)

    parser.add_argument("-v", "--version", action="version", version="1.0.0", help="show the version number and exit")
                       
    return parser.parse_args()


def str2bool(b):
    if isinstance(b, bool):
       return b
    if b.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif b.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')        

if __name__ == "__main__":
    main(parseArgs())



'''
Extract and return the blast alignment paths
'''

import os, sys, re
from collections import defaultdict

from tipp3 import get_logger
from tipp3.helpers.alignment_tools import ExtendedAlignment

_LOG = get_logger(__name__)

global complement_map
complement_map = {'A': 'T', 'T': 'A', 'C': 'G', 'G': 'C'}

'''
Reverse-complement given sequences. Retain characters that are not {A, T, C, G}
'''
def reverseComplement(seq):
    try:
        char_list = [complement_map[c] for c in seq]
    except KeyError:
        char_list = [complement_map[c] if c in complement_map else c for c in seq]
    new_seq = ''.join(char_list)
    return new_seq[::-1]

'''
Given an aligned string (represented by upper/lower case letters and gaps,
return a condensed version that have the lowercase letters from both sides
compressed to front/back.
'''
def compressInsertions(seq):
    p = re.compile(r'[A-Z]+')
    alns = [(m.start(), m.end()) for m in p.finditer(seq)]
    # do not perform such task if there is no aligned column at all
    if len(alns) == 0:
        return seq

    # first occurrence of aligned position defines the back of front search space
    # i.e., [start, end)
    f_start, f_end = 0, alns[0][0]
    f_len = f_end - f_start

    # last occurrence of aligned position defines the front of the back search space
    b_start, b_end = alns[-1][1], len(seq)
    b_len = b_end - b_start

    # simplest way of compression: remove all gaps and add them back
    f_str_ins = seq[f_start:f_end].replace('-', '')
    f_len_ins = len(f_str_ins)
    f_str = f_str_ins + '-' * (f_len - f_len_ins)

    b_str_ins = seq[b_start:b_end].replace('-', '')
    b_len_ins = len(b_str_ins)
    b_str = '-' * (b_len - b_len_ins) + b_str_ins

    # combine the compressed front/back with the remaining sequence
    combined = f_str + seq[f_end:b_start] + b_str

    return combined

'''
Obtain retained columns of a target sequence
'''
def getRetainedColumns(aln, target_keys):
    target_to_retained_columns = {}
    aln_len = aln.sequence_length()
    
    # we only care about where each character appears in the backbone columns
    for target in target_keys:
        retained_columns = defaultdict(int)
        idx = 0
        for i in range(aln_len):
            if aln[target][i] != '-':
                retained_columns[idx] = i
                idx += 1
        target_to_retained_columns[target] = retained_columns
    return target_to_retained_columns

'''
Obtain aligned columns of query seq to target
'''
def getAlignedColumns(taxon, seq_len, aln_seq, target_seq, target_start):
    # obtain aligned columns of the query sequence (to the target sequence)
    qidx = 0; tidx = target_start
    aligned_columns = [-1] * seq_len
    for i in range(len(aln_seq)):
        # if current is a deletion, increment tidx
        if aln_seq[i] == '-':
            tidx += 1
            continue
        else:
            # current is an insertion, do not increment tidx and map to -1
            if target_seq[i] == '-':
                aligned_columns[qidx] = -1
                qidx += 1
            # current is a match, increment both indexes 
            else:
                # current 
                aligned_columns[qidx] = tidx
                qidx += 1
                tidx += 1
    return aligned_columns

#'''
#Obtain alignment_graph and backtrace to find the query merged alignment
#'''
#def getBacktrace(aligned_columns, retained_columns):
#    # target start marks from where blast considers the alignment of the target sequence
#    alignment_graph = [[0 for _j in range(backbone_length + 1)] for _i in range(seq_len + 1)]
#    weights = defaultdict(int)
#    min_col_ind, max_col_ind = backbone_length + 1, -1
#
#    for i in range(len(aligned_columns)):
#        subset_col = aligned_columns[i]
#        # ignore insertions
#        if subset_col == -1:
#            continue
#        j = retained_columns[subset_col]
#        weights[(i, j)] = 1   # meaning query[i] is aligned to backbone position j
#        min_col_ind = min(min_col_ind, j)
#        max_col_ind = max(max_col_ind, j)
#    # print(weights)
#
#    backtrace = [[0 for _j in range(backbone_length + 1)] for _i in range(seq_len + 1)]
#    for i in range(0, seq_len + 1):
#        for j in range(min_col_ind, max_col_ind + 2):
#            if i == 0 or j == min_col_ind:
#                alignment_graph[i][j] = 0   # on the edge of the alignment graph
#                continue
#            cur_max = 0; cur_bt = 0
#            w = weights.get((i-1, j-1), 0)
#            values = [alignment_graph[i-1][j-1] + w,
#                      alignment_graph[i-1][j],
#                      alignment_graph[i][j-1]]
#            for _ind, val in enumerate(values):
#                if _ind == 0 and w <= 0:
#                    cur_bt = 1    # in the case position i-1 in query is not mapped to j-1 in backbone, prefer going "upward"
#                    continue
#                if val > cur_max:
#                    cur_max = val
#                    cur_bt = _ind
#            alignment_graph[i][j] = cur_max
#            backtrace[i][j] = cur_bt
#    return backtrace, min_col_ind, max_col_ind

#'''
#Get combined aligned query sequence from backtrace
#'''
#def getCombinedSequence(seq, seq_len, backtrace, min_col_ind, max_col_ind):
#    # retrieve results from backtrace
#    result = []; i, j = seq_len, max_col_ind + 1
#    while i > 0 and j > min_col_ind:
#        bt = backtrace[i][j]
#        if bt == 0:
#            result.append(seq[i-1].upper())
#            i -= 1; j -= 1
#        elif bt == 1:
#            # query "insertion", lowercase
#            result.append(seq[i-1].lower())
#            i -= 1
#        elif bt == 2:
#            # query "deletion", gap
#            result.append('-')
#            j -= 1
#        else:
#            raise ValueError
#    while i > 0:
#        result.append(seq[i-1].lower()); i -= 1
#    while j > min_col_ind:
#        result.append('-'); j -= 1
#
#    # reverse result to be in correct order
#    result = result[::-1]
#
#    # append '-' to start and end of results to fill up the alignment
#    result = ['-'] * min_col_ind + result + ['-'] * (backbone_length - max_col_ind - 1)
#
#    # result
#    combined = compressInsertions(''.join(result))
#    return combined

'''
Convert combined sequence to an Extended Alignment object
'''
def getQueryExtendedAlignment(taxon, combined):
    # get an extended alignment object with just the query sequence and its updatd indexes
    query = ExtendedAlignment([])
    query[taxon] = combined; query._reset_col_names()
    insertion = -1; regular = 0
    for i in range(len(combined)):
        if combined[i].islower():
            query._col_labels[i] = insertion; insertion -= 1
        else:
            query._col_labels[i] = regular; regular += 1
    return query

'''
Directly construct the query alignment to backbone by setting characters based on retained_columns
'''
def constructCombined(seq, backbone_length, aligned_columns, retained_columns):
    ret = ['-'] * backbone_length   # initialize as gaps, later modify to be characters from seq
    for i in range(len(aligned_columns)):
        target_col = aligned_columns[i]
        if target_col == -1:
            continue
        # assign seq[i] to ret[retained_columns[target_col]]
        ret[retained_columns[target_col]] = seq[i]
    return ''.join(ret)

def extractionRunner(marker, inpath, bbpath, outpath):
    # get mapped queries to marker gene
    aln = ExtendedAlignment([])
    aln.read_file_object(bbpath)
    backbone_length = aln.sequence_length()

    mapped = defaultdict(list)
    with open(inpath, 'r') as f:
        lines = f.read().strip().split('\n')
        for i in range(0, len(lines), 2):
            taxon = lines[i].split('>')[1]
            items = eval(lines[i+1])
            mapped[taxon] = items

    # compute retained columns just for once for the target marker gene sequences that appear
    target_keys = set()
    for taxon in mapped.keys():
        target_keys.add(mapped[taxon][0][0])
    target_to_retained_columns = getRetainedColumns(aln, target_keys)

    #for taxon in tqdm(mapped.keys(), total=len(mapped), **tqdm_styles):
    for taxon in mapped.keys():
        aln_seq = mapped[taxon][3]
        target = mapped[taxon][0][0]
        target_seq = mapped[taxon][6]

        seq = aln_seq.replace('-', '')
        seq_len = len(seq)
        # deal with reverse-complement
        # target sequence is reverse-complemented
        #   - we need to restore that
        #   - and make the query reverse-complemented
        if mapped[taxon][4] > mapped[taxon][5]:
            target_start = mapped[taxon][5] - 1
            aln_seq = reverseComplement(aln_seq)
            target_seq = reverseComplement(target_seq)
            seq = reverseComplement(seq)
        else:
            target_start = mapped[taxon][4] - 1

        # obtain aligned columns of taxon
        aligned_columns = getAlignedColumns(taxon, seq_len, aln_seq, 
                target_seq, target_start)
        # obtain retained columns of target
        retained_columns = target_to_retained_columns[target]
        # directly construct the aligned query sequence by adding gaps where necessary and ignoring insertion characters
        # this will be the masked version
        combined = constructCombined(seq, backbone_length, aligned_columns, retained_columns)
        aln[taxon] = combined
    # write a masked version to local (masking lower case letters)
    # NOTE: only retaining query keys to save space
    subaln = aln.sub_alignment(list(mapped.keys()))
    subaln.write(outpath, 'FASTA')

    ## write a query-only version
    #subaln = aln.sub_alignment(list(mapped.keys()))
    #subaln.write(outqueriespath, 'FASTA')

    ## write a backbone-only version
    #bbset = set(aln.keys()).difference(set(mapped.keys()))
    #bbaln = aln.sub_alignment(list(bbset))
    #bbaln.write(outbackbonepath, 'FASTA')

'''
entry point for extracting BLAST alignment
'''
def extractBlastAlignment(refpkg, workdir, query_blast_paths):
    _LOG.info("Started extracting query pairwise alignments from BLAST output") 

    # each marker's BLAST result path is determined in query_blast_paths
    alignment_outdir = workdir + '/query_alignments'
    if len(query_blast_paths) == 0:
        _LOG.warning('No query reads are assigned to any marker '
        'gene according BLAST.')
        return None

    ret = {}
    for marker, inpath in query_blast_paths.items():
        bbpath = refpkg[marker]['alignment']
        outdir = os.path.join(alignment_outdir, marker)
        if not os.path.isdir(outdir):
            os.makedirs(outdir)
        # NOTE: only keep the query alignments (backbone part is unnecessary)
        outpath = os.path.join(outdir, 'est.aln.masked.queries.fasta')
        ret[marker] = outpath

        if os.path.exists(outpath) and os.stat(outpath).st_size > 0:
            _LOG.info(f"Skipping {marker}, found existing alignment: {outpath}")
            continue

        #outqueriespath = os.path.join(outdir, 'est.aln.masked.queries.fasta')
        #outbackbonepath = os.path.join(outdir, 'est.aln.masked.backbone.fasta')
        extractionRunner(marker, inpath, bbpath, outpath)
        _LOG.debug('Writing BLAST alignment for marker gene {} to {}'.format(
            marker, outpath))
    return ret

    #inpath = sys.argv[1]    # categorized by marker gene
    #bbpath = sys.argv[2]    # marker gene backbone alignment
    #outpath = sys.argv[3]   # output with query alignments merged
    #outqueriespath = sys.argv[4]    # output with only queries
    #outbackbonepath = sys.argv[5]    # output with only backbone sequences
    #runner(inpath, bbpath, outpath, outqueriespath, outbackbonepath)




'''
Place reads to marker gene taxonomic tree using a placement method
'''

import os, time, shutil
from tipp3 import get_logger
from tipp3.configs import Configs
from tipp3.jobs import BscamppJob, PplacerTaxtasticJob

_LOG = get_logger(__name__)


'''
Place query reads to their assigned marker gene taxonomic trees using the
user-defined phylogenetic placement method
'''
def queryPlacement(refpkg, query_alignment_paths):
    _LOG.info(f"Started placement method: {Configs.placement_method}")

    query_placement_paths = {}
    additional_kwargs = {}
    try:
        #additional_kwargs = Configs.BSCAMPP.__dict__
        additional_kwargs = getattr(Configs, Configs.placement_method).__dict__
    except (AttributeError, ValueError) as e:
        pass
    _LOG.info("(From main.config) parameter options for " \
            f"{Configs.placement_method}: {additional_kwargs}")

    # failsafe when no alignments are given
    if len(query_alignment_paths) == 0:
        _LOG.warning("No alignments are given for placements, returning...")
        return query_placement_paths
    
    # removing temporary files that we do not care
    temp_folders = ['tmp0']

    idx = 0; total = len(query_alignment_paths) 
    for marker, query_alignment_path in query_alignment_paths.items():
        _LOG.info("Placing aligned query reads for " \
                f"{marker}") #from {query_alignment_path}")
        idx += 1
        placement_dir = os.path.join(Configs.outdir, 'query_placements',
                marker)
        if not os.path.isdir(placement_dir):
            os.makedirs(placement_dir)

        # detect results from previous run, if exists return
        detect_path = os.path.join(placement_dir, 'placement.jplace')
        if os.path.exists(detect_path) and os.stat(detect_path).st_size > 0:
            _LOG.info(f"Found existing placement: {detect_path}")
            query_placement_paths[marker] = detect_path
            continue

        # identify job type - different placement methods
        # pplacer-taxtastic
        if Configs.placement_method == 'pplacer-taxtastic':
            placement_job = PplacerTaxtasticJob(path=Configs.pplacer_path,
                    query_alignment_path=query_alignment_path,
                    refpkg_path=refpkg[marker]['path'],
                    outdir=placement_dir,
                    num_cpus=Configs.num_cpus,
                    **additional_kwargs)

        # default: bscampp Job
        else:
            backbone_alignment_path = refpkg[marker]['alignment']

            placement_job = BscamppJob(path=Configs.bscampp_path,
                    query_alignment_path=query_alignment_path,
                    backbone_alignment_path=backbone_alignment_path,
                    backbone_tree_path=refpkg[marker]['additional-raxml-br-tree'],
                    tree_model_path=refpkg[marker]['additional-raxml-model-file'],
                    outdir=placement_dir,
                    num_cpus=Configs.num_cpus,
                    **additional_kwargs) 
        placement_path = placement_job.run(logging=True)
        query_placement_paths[marker] = placement_path

        # clean up some temp files
        for tf in temp_folders:
            toremove = os.path.join(placement_dir, tf)
            if os.path.exists(toremove):
                shutil.rmtree(toremove)
        _LOG.info(f"({idx}/{total}) Finished placing aligned query reads for {marker}")
    return query_placement_paths



import os, sys, shutil
try:
    import configparser
except ImportError:
    import ConfigParser as configparser
from argparse import ArgumentParser, Namespace
from platform import platform

def find_main_config(homepath):
    with open(homepath, 'r') as f:
        _root_dir = f.read().strip()
        main_config_path = os.path.join(_root_dir, 'main.config')
        if os.path.exists(main_config_path):
            return _root_dir, main_config_path
        else:
            return None, None

'''
first time run, need user to initialize the main.config
if it is not installed through github (i.e., python setup.py config)
will be needed if installed through pip/pypi
'''
def init_config_file(homepath, rerun=False, prioritize_user_software=True):
    # read from sys.argv to find if "-y" or "--bypass-setup" exists
    args = sys.argv[1:]
    bypass_setup = True
    #if '-y' in args or '--bypass-setup' in args:
    #    bypass_setup = True

    # initialize a home.path that points to local user main.config
    # if it exists then pass on
    if not rerun:
        if os.path.exists(homepath):
            # detecting old home.path file based on creation time
            if os.stat(homepath).st_mtime >= os.stat(__file__).st_mtime:
                _root_dir, main_config_path = find_main_config(homepath)
                if _root_dir is None:
                    print('home.path exists but main.config missing, regenerating...')
                else:
                    return _root_dir, main_config_path
            else:
                print('Found old home.path and regenerating...')
                os.remove(homepath)
        else:
            print('Cannot find home.path: {}'.format(homepath))
    else:
        print('Re-initializing the config file...')

    # install to user local directory
    # bypassing the setup step to directly use the default path
    _root_dir = ''
    if not bypass_setup:
        _root_dir = input('Create main.config file at [default: ~/.tipp3/]: ')

    if _root_dir == '':
        _root_dir = os.path.expanduser('~/.tipp3')
    else:
        _root_dir = os.path.abspath(_root_dir)
    main_config_path = os.path.join(_root_dir, 'main.config')
    print('Initializing main configuration file: {}...'.format(main_config_path))

    # write to local for installation to system
    # will read in during runs to find the main.config file
    if not os.path.isdir(_root_dir):
        os.mkdir(_root_dir)
    with open(homepath, 'w') as f:
        f.write(_root_dir)

    # create main.config file at configfile using default.config
    _config_path = os.path.join(os.path.dirname(__file__), 'default.config')
    config_defaults = []
    cparser = configparser.ConfigParser()
    cparser.optionxform = str
    assert os.path.exists('{}'.format(_config_path)), \
            "default config file {} missing! Please redownload from Github\n".format(
                    _config_path)

    if os.path.exists(main_config_path):
        print('Main configuration file {} exists...'.format(main_config_path))
        print('Overwriting existing configuration file...')

    print('\n')
    # initialize main config file using default config file
    default_config = configparser.ConfigParser()
    with open(_config_path, 'r') as f:
        default_config.read_file(f)
    for section in default_config.sections():
        cparser.add_section(section)
        for k, v in default_config[section].items():
            cparser.set(section, k, v)

    # if platform is linux then we just copy the default config file
    # as main.config
    platform_name = platform()
    tools_dir = os.path.join(os.path.dirname(__file__), 'tools')
    set_sections = ['basic']

    # default path to pplacer, BSCAMPP, and TIPPJsonMerger
    cparser.set('basic', 'pplacer_path',
            os.path.join(tools_dir, 'pplacer', 'pplacer'))
    cparser.set('basic', 'bscampp_path',
            os.path.join(tools_dir, 'bscampp', 'EPA-ng-BSCAMPP.py'))
    cparser.set('basic', 'tippjsonmerger_path',
            os.path.join(tools_dir, 'merge', 'tippJsonMerger.jar'))

    if 'macos' not in platform_name.lower():
        print('System is {}, using default config as main.config...'.format(
            platform_name))
        # use existing binaries from MAGUS subfolder (reduce redundancy of
        # duplicated binaries)
        #for _section in set_sections:
        #    pass
    else:
        raise NotImplementedError("Software haven't tuned for macOS yet.")
        #if 'x86' not in platform_name:
        #    print('Warning: system is not using x86 architecture.',
        #            'Some softwares such as FastTreeMP need to be',
        #            'self-provided. See {} [Basic] '.format(_config_path),
        #            'section for more information.')
        #print("System is {}, reconfiguring main.config...".format(platform_name))

        ## configure MAGUS to use macOS compatible executables
        #binaries = os.listdir(tools_dir + '/macOS')
        #for binary in binaries:
        #    path = os.path.join(tools_dir, 'macOS', binary)
        #    #path = os.path.join(_macOS_dir, binary)
        #    for _section in set_sections:
        #        if 'FastTreeMP' in path:
        #            cparser.set(_section, 'fasttreepath', path)
        #                    #self.copy_tool_to_lib('FastTreeMP', path))
        #        else:
        #            cparser.set(_section, '{}path'.format(binary), path)
        #                    #self.copy_tool_to_lib(binary, path))

    # binaries from the user's environment will be used in priority
    # if they exist
    if prioritize_user_software:
        print('Detecting existing software from the user\'s environment...')
        software = ['witch.py', 'pplacer', 'blastn']
        print('\tDetected:\n')
        for soft in software:
            if shutil.which(soft):
                print('\t{}: {}'.format(soft, shutil.which(soft)))
                for _section in set_sections:
                    if soft == 'witch.py':
                        cparser.set(_section, 'witch_path',
                                shutil.which(soft))
                    else:
                        cparser.set(_section, '{}_path'.format(soft),
                                shutil.which(soft))

    with open(main_config_path, 'w') as f:
        cparser.write(f)
    print('\n(Done) main.config written to {}'.format(main_config_path))
    print('If you would like to make manual changes, please directly edit {}'.format(
        main_config_path))
    # DO NOT EXIT; can start running TIPP3 with any given commands now
    return _root_dir, main_config_path



import psutil, os
import argparse

tqdm_styles = {
        'desc': '\tRunning...', 'ascii': False,
        'ncols': 80,
        #'disable': True,
        #'colour': 'green',
        'mininterval': 0.5
        }

# return memory usage of python process by MB
def memoryUsage():
    process = psutil.Process(os.getpid())
    mem = process.memory_info().rss / float(2 ** 20)
    return mem

# reformat argparse help text formatting
class SmartHelpFormatter(argparse.RawDescriptionHelpFormatter):
    def add_text(self, text):
        if text is not None:
            text = text.replace("\\n", "\n").replace("\\t", "\t")
        super().add_text(text)
    def _split_lines(self, text, width):
        if '\n' in text:
            temp = text.split('\n')
            ret = []
            for _splice in [argparse.RawDescriptionHelpFormatter._split_lines(self, x, width)
                    for x in temp]:
                ret.extend(_splice)
            return ret
        return argparse.RawDescriptionHelpFormatter._split_lines(self, text, width)
